from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client


def _format_number(n):
    # This is to mimic the following JS code:
    # n.toLocaleString(undefined, {maximumFractionDigits: 4})
    # It's possible different locales will display it differently in JS, but using n
    # instead of , in the format string lead to exponent view, which is incorrect.
    # Reference: https://stackoverflow.com/questions/1823058/how-to-print-a-number-using-commas-as-thousands-separators
    rounded_number = round(n, ndigits=4)
    return f"{rounded_number:,}"


def _format_string(n):
    # This is to mimic String in JS code.
    if isinstance(n, bool):
        return str(n).lower()
    return str(n)


def _get_numeric_buckets(segment_data: Dict) -> List[Dict]:
    values = segment_data["values"]
    result = []
    for i in range(len(values) - 1):
        value = values[i]
        next_value = values[i + 1]
        name = f"{_format_number(value)} - {_format_number(next_value)}"
        result.append({"value": value, "nextValue": next_value, "name": name})

    return result


def _get_non_numeric_buckets(segment_data: Dict) -> List[Dict]:
    values = segment_data["values"]
    return [{"value": value, "name": _format_string(value)} for value in values]


class Segment(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)
        self.extended_data = None

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.field_name = data["field_name"]
        self.values = data["values"]
        self.terms = data["terms_values"]

    def __update_extended_data(self):
        response = self.client.send_request(
            f"/data-segments/{self.id}",
            "GET",
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)
        self.extended_data = response.json()

    def get_extended_data(self) -> Dict:
        if self.extended_data is None:
            self.__update_extended_data()
        return self.extended_data

    def get_widget_values(self) -> List[Any]:
        if self.extended_data is None:
            self.__update_extended_data()

        if (
            self.extended_data["is_custom_segment"]
            or self.extended_data["field"]["type"] != "numeric"
        ):
            return self.values
        return self.values[:-1]

    def to_widget(self) -> Dict:
        if self.extended_data is None:
            self.__update_extended_data()

        segment_data = {
            "id": self.id,
            "name": self.name,
            "createdAt": self.raw_data["created_at"],
            "isGlobalFilter": self.extended_data["create_composite_segments"],
            "values": self.extended_data["values"],
        }

        if self.extended_data["is_custom_segment"]:
            segment_data["type"] = "custom"
            segment_data["termsValues"] = self.extended_data["terms_values"]
        else:
            segment_data["type"] = "auto"
            segment_data["fieldId"] = self.extended_data["field"]["id"]
            segment_data["fieldType"] = self.extended_data["field"]["type"]
            if segment_data["fieldType"] == "numeric":
                segment_data["buckets"] = _get_numeric_buckets(self.extended_data)

        if "buckets" not in segment_data:
            segment_data["buckets"] = _get_non_numeric_buckets(self.extended_data)

        return segment_data

    @classmethod
    def get_all(
        cls, client: Client, model_id: Optional[str] = None, manual_only: bool = True
    ) -> List["Segment"]:
        url_params = {}
        if manual_only:
            url_params["is_manually_created"] = "true"
        if model_id is not None:
            url_params["model_id"] = model_id
        encoded_url_params = urlencode(url_params)
        response = client.send_request(
            f"/data-segments{f'?{encoded_url_params}' if len(url_params) > 0 else f''}",
            "GET",
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        model_id: str,
        field_name: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        terms: Optional[List[Tuple[str, str]]] = None,
        is_global_filter: bool = False,
    ) -> "Segment":
        segment_data = {}
        if terms is not None:
            segment_data["terms_values"] = terms
        else:
            segment_data["field_name"] = field_name
            segment_data["values"] = values
        segment_data["is_global_filter"] = is_global_filter
        response = client.send_request(
            "/data-segments",
            "POST",
            {"name": name, "model_id": model_id, **segment_data},
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Segment":
        response = client.send_request(f"/data-segments/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(
        self,
        name: Optional[str] = None,
        field_name: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        terms: Optional[List[Tuple[str, str]]] = None,
        is_global_filter: Optional[bool] = None,
        **kwargs,
    ):
        args = {}
        if name is not None:
            args["name"] = name
        if field_name is not None:
            args["field_name"] = field_name
        if values is not None:
            args["values"] = values
        if terms is not None:
            args["terms_values"] = terms
        if is_global_filter is not None:
            args["is_global_filter"] = is_global_filter
        response = self.client.send_request(
            f"/data-segments/{self.id}",
            "PUT",
            args,
        )
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/data-segments/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/data-segments/{id}", "DELETE")
        client.assert_response(response)



from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from importlib.metadata import version as get_version
import json
import pathlib
import string
import traceback
from typing import Any, Dict, List, Optional, Tuple

from requests import Session

from aporia.as_code.resources.base import (
    BaseResource,
    CompareStatus,
    NonDeletableResourceException,
)
from aporia.as_code.resources.custom_metrics import CustomMetric
from aporia.as_code.resources.data_source import DataSource
from aporia.as_code.resources.dataset import Dataset
from aporia.as_code.resources.model import Model
from aporia.as_code.resources.monitor import Monitor
from aporia.as_code.resources.segment import Segment
from aporia.as_code.resources.version import Version
from aporia.sdk.aporia_api import Aporia
from aporia.sdk.aporia_api import BaseAporiaResource as SDKBaseAporiaResource
from aporia.sdk.client import Client, EntityNotFoundException
from aporia.sdk.custom_metrics import CustomMetric as SDKCustomMetric
from aporia.sdk.datasets import Dataset as SDKDataset
from aporia.sdk.metrics import MetricType
from aporia.sdk.models import Model as SDKModel
from aporia.sdk.models import ModelColor
from aporia.sdk.monitors import (
    ActionType,
    infer_detection_method,
    LogicEvaluationType,
    MessagingIntegrationType,
)
from aporia.sdk.monitors import Monitor as SDKMonitor
from aporia.sdk.monitors import (
    MonitorType,
    parse_monitor_configuration,
    PreConditionType,
)
from aporia.sdk.segments import Segment as SDKSegment
from aporia.sdk.versions import Version as SDKVersion

sdk_version = get_version("aporia")


class StackAction(Enum):
    CREATE = "create"
    DELETE = "delete"
    PREVIEW = "preview"
    REFRESH = "refresh"
    DIFF = "diff"


class ResourceStatus(Enum):
    CREATED = "created"  # Found in config file
    MISSING = "missing"  # Not found in config file, create
    DEPRECATED = "deprecated"  # Found in config file, definition changed
    # TODO: Support that
    # UNWANTED = "unwanted"  # Found in config file, not in resources


@dataclass
class Resource:
    resource: BaseResource
    depth: int
    status: ResourceStatus
    id: Optional[str] = None


RESOURCE_TYPE_TO_RESOURCE_CLASS = {
    CustomMetric.__name__: CustomMetric,
    DataSource.__name__: DataSource,
    Dataset.__name__: Dataset,
    Model.__name__: Model,
    Monitor.__name__: Monitor,
    Segment.__name__: Segment,
    Version.__name__: Version,
}


class Stack:
    def __init__(
        self,
        token: str,
        account: str,
        workspace: str = "default-workspace",
        base_url: str = "https://platform.aporia.com",
        requests_session: Optional[Session] = None,
    ):
        self._resources: OrderedDict[str, Resource] = OrderedDict()
        self._output = None
        self._base_url = base_url
        self._client = Client(
            base_url=f"{base_url}/api/v1/{account}/{workspace}",
            token=token,
            requests_session=requests_session,
        )
        self._delete_resources: OrderedDict[str, BaseResource] = OrderedDict()
        self._aporia_api = Aporia(
            token=token,
            account_name=account,
            workspace_name=workspace,
            base_url=base_url,
            requests_session=requests_session,
        )

    def add(self, resource: BaseResource, depth=0, prefix=None):
        unique_id = (prefix or "") + resource.name

        if unique_id in self._resources.keys():
            raise Exception(f"Found duplicate resource name: {unique_id}")

        self._resources[unique_id] = Resource(
            resource=resource, depth=depth, status=ResourceStatus.MISSING
        )
        if hasattr(resource, "sub_resources"):
            for sub_resource, _ in resource.sub_resources:
                self.add(
                    sub_resource,
                    depth=depth + 1,
                    prefix=f"{prefix or ''}{resource.name}/",
                )

    def _call_argsetters(self, resource, resource_data):
        if hasattr(resource, "sub_resources"):
            sub_resources = resource.sub_resources
            for sub_resource, argsetter in sub_resources:
                argsetter(resource_data, sub_resource)
        if hasattr(resource, "dependants"):
            dependants = resource.dependants
            for sub_resource, argsetter in dependants:
                argsetter(resource_data, sub_resource)

    def _preview(self, delete: bool = False):
        if delete:
            RESOURCE_STATUS_TO_PREVIEW_ACTION = {
                ResourceStatus.CREATED: "delete",
                ResourceStatus.MISSING: "create",
                ResourceStatus.DEPRECATED: "delete",
            }
            for resource_entry in self._resources.values():
                if resource_entry.status is ResourceStatus.MISSING:
                    continue
                print(
                    f"|-{'----' * resource_entry.depth}{resource_entry.resource.name} - {RESOURCE_STATUS_TO_PREVIEW_ACTION[resource_entry.status]}"
                )
        else:
            RESOURCE_STATUS_TO_PREVIEW_ACTION = {
                ResourceStatus.CREATED: "pass",
                ResourceStatus.MISSING: "create",
                ResourceStatus.DEPRECATED: "update",
            }
            for resource_entry in self._resources.values():
                print(
                    f"|-{'----' * resource_entry.depth}{resource_entry.resource.name} - {RESOURCE_STATUS_TO_PREVIEW_ACTION[resource_entry.status]}"
                )
        if len(self._delete_resources) > 0:
            print("Removed resources:")
        for unique_id in self._delete_resources.keys():
            resource_parts = unique_id.split("/")
            depth = len(resource_parts) - 1
            name = resource_parts[-1]
            print(f"|-{'----' * depth}{name} - delete")

    def _get_diff_config(self, output_format: str = "print"):
        diffs = {}
        for unique_id, resource_entry in self._resources.items():
            if resource_entry.status is not ResourceStatus.DEPRECATED:
                continue
            resource_data = self._output[unique_id]
            resource_diff = resource_entry.resource.get_diff(resource_data=resource_data)
            diffs[unique_id] = {}
            for config_name, (old, new) in resource_diff.items():
                diffs[unique_id][config_name] = {"old": old, "new": new}
        if output_format == "json":
            print(json.dumps(diffs))
        elif output_format == "print":
            for unique_id, diffed_configs in diffs.items():
                print(f"Resource '{unique_id}' changed. Changes:")
                for config_name, config_diffs in diffed_configs.items():
                    print(
                        f"    Configuration '{config_name}' changed.\n        Old: {config_diffs['old']}\n        New: {config_diffs['new']}"
                    )

    def __pop_deleted_sub_resources(self, non_deleteable_resources: List[str], unique_id):
        resource_entry = self._resources[unique_id]
        if hasattr(resource_entry.resource, "sub_resources"):
            for resource, _ in resource_entry.resource.sub_resources:
                sub_resource_unique_id = f"{unique_id}/{resource.name}"
                if sub_resource_unique_id in non_deleteable_resources:
                    self.__pop_deleted_sub_resources(
                        non_deleteable_resources=non_deleteable_resources,
                        unique_id=sub_resource_unique_id,
                    )
                    self._output.pop(sub_resource_unique_id)
                elif sub_resource_unique_id in self._output.keys():
                    raise RuntimeError(
                        f"Resource deletion missed for {sub_resource_unique_id}. Stack potentially corrupted."
                    )

    def _create(self, debug: bool = False) -> Tuple[Dict, Optional[Exception]]:
        output = self._output or {}
        error = None
        try:
            # Create/update resources
            for (
                unique_id,
                resource_entry,
            ) in self._resources.items():
                if resource_entry.status is ResourceStatus.CREATED:
                    continue
                if hasattr(resource_entry.resource, "deferred_load"):
                    resource_entry.resource.deferred_load(self._client, resource_entry.resource)

                if resource_entry.status is ResourceStatus.MISSING:
                    if debug:
                        print(f"Attempting to create: {unique_id}")
                    _, resource_data = resource_entry.resource.create(client=self._client)
                elif resource_entry.status is ResourceStatus.DEPRECATED:
                    resource_data = resource_entry.resource.update(
                        client=self._client, id=self._output[unique_id]["id"]
                    )
                else:
                    raise Exception("Unexpected resource status")
                self._call_argsetters(resource_entry.resource, resource_data)
                output[unique_id] = resource_data
                output[unique_id]["_resource_type"] = type(resource_entry.resource).__name__
                resource_entry.status = ResourceStatus.CREATED

            # Delete resources
            non_deleteable_resources = []
            for unique_id, resource_class in list(self._delete_resources.items())[::-1]:
                resource_data = self._output[unique_id]
                if debug:
                    print(f"Attempting to delete: {unique_id}")
                try:
                    resource_class.delete(client=self._client, id=resource_data["id"])
                except NonDeletableResourceException:
                    non_deleteable_resources.append(unique_id)
                    continue
                for non_deleteable_resource in [*non_deleteable_resources]:
                    if non_deleteable_resource.startswith(f"{unique_id}/"):
                        self._output.pop(non_deleteable_resource)
                        non_deleteable_resources.remove(non_deleteable_resource)
                self._output.pop(unique_id)
        except Exception as e:
            error = e

        return output, error

    def _destroy(self):
        # Read from storage
        # Destroy all resources
        # TODO: Consider adding all this logic into _create, just with setting resource.action to delete
        error = None
        if self._output is None:
            raise Exception("Stack not created")
        try:
            non_deleteable_resources = []
            for unique_id, resource_entry in list(self._resources.items())[::-1]:
                if unique_id in self._output.keys():
                    resource_data = self._output[unique_id]
                    try:
                        resource_entry.resource.delete(client=self._client, id=resource_data["id"])
                    except NonDeletableResourceException:
                        non_deleteable_resources.append(unique_id)
                        continue
                    self.__pop_deleted_sub_resources(non_deleteable_resources, unique_id)
                    self._output.pop(unique_id)
            for unique_id, resource_class in list(self._delete_resources.items())[::-1]:
                resource_data = self._output[unique_id]
                try:
                    resource_class.delete(client=self._client, id=resource_data["id"])
                except NonDeletableResourceException:
                    non_deleteable_resources.append(unique_id)
                    continue
                self.__pop_deleted_sub_resources(non_deleteable_resources, unique_id)
                self._output.pop(unique_id)
        except Exception as e:
            error = e

        return error

    def _diff(self):
        for unique_id, resource_entry in self._resources.items():
            if unique_id in self._output.keys():
                # Resource already exists. Call argsetters
                # TODO: Add support for edit
                resource_data = self._output[unique_id]
                self._call_argsetters(resource_entry.resource, resource_data)
                compare_status = self._resources[unique_id].resource.compare(
                    resource_data=resource_data
                )
                if compare_status is CompareStatus.SAME:
                    self._resources[unique_id].status = ResourceStatus.CREATED
                elif compare_status is CompareStatus.UPDATEABLE:
                    self._resources[unique_id].status = ResourceStatus.DEPRECATED
                else:
                    raise Exception(
                        f"Configuration for {unique_id} was edited but can't be updated!"
                    )
                continue

        for unique_id, resource_data in self._output.items():
            if unique_id not in self._resources.keys():
                self._delete_resources[unique_id] = RESOURCE_TYPE_TO_RESOURCE_CLASS[
                    resource_data["_resource_type"]
                ]

    def __resource_name_to_python_name(self, name: str) -> str:
        result = "".join(
            [c if c in (string.ascii_letters + string.digits + "_") else "_" for c in name]
        ).lower()
        if result[0] in string.digits:
            result = f"_{result}"
        return result

    def __create_python_representation(
        self,
        resource: SDKBaseAporiaResource,
        index: int,
        version_id_to_object: Optional[Dict[str, str]] = None,
        segment_id_to_object: Optional[Dict[str, str]] = None,
        custom_metric_id_to_object: Optional[Dict[str, str]] = None,
    ) -> Tuple[str, str]:
        if isinstance(resource, SDKSegment):
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Segment(
    "{resource.name}",
"""
            if resource.raw_data["field_name"] is not None:
                code += f"""\tfield="{resource.raw_data["field_name"]}",\n"""
                code += f"""\tvalues={resource.raw_data["values"]}\n"""
            else:
                code += f"""\tterms={resource.raw_data["terms_values"]}\t"""
            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKCustomMetric):
            return (
                f'''{self.__resource_name_to_python_name(resource.name)} = aporia.CustomMetric(
    "{resource.name}",
    code="""{resource.raw_data["code"]}""",
)''',
                self.__resource_name_to_python_name(resource.name),
            )
        elif isinstance(resource, SDKDataset):
            schema = resource.raw_data["schema"]
            raw_inputs = {
                raw_input["name"]: raw_input["type"] for raw_input in schema["raw_inputs"]
            }
            features = {feature["name"]: feature["type"] for feature in schema["features"]}
            predictions = {
                prediction["name"]: prediction["type"] for prediction in schema["predictions"]
            }
            actuals = {actual["name"]: actual["type"] for actual in schema["actuals"]}
            actual_mappings = {
                actual["name"]: actual["properties"]["prediction"] for actual in schema["actuals"]
            }
            code = f"""{resource.raw_data["stage"]}_dataset_{index} = aporia.Dataset(
    "{resource.raw_data["stage"]}",
    type="{resource.raw_data["stage"]}",
    data_source_id="{resource.raw_data["data_source_id"]}",
    connection_data={resource.raw_data["config"]},
"""
            if schema["id_column"] is not None:
                code += f"""\tid_column="{schema['id_column']}",\n"""
            if schema["timestamp_column"] is not None:
                code += f"""\ttimestamp_column="{schema['timestamp_column']}",\n"""
            if len(raw_inputs) > 0:
                code += f"\traw_inputs={raw_inputs},\n"
            if len(features) > 0:
                code += f"\tfeatures={features},\n"
            if len(predictions) > 0:
                code += f"\tpredictions={predictions},\n"
            if len(actuals) > 0:
                code += f"\tactuals={actuals},\n"
            if len(actual_mappings) > 0:
                code += f"\tactual_mappings={actual_mappings},\n"
            code += ")"
            return code, f'{resource.raw_data["stage"]}_dataset_{index}'
        elif isinstance(resource, SDKVersion):
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Version(
    "{resource.name}",
"""
            if resource.raw_data["serving_dataset"] is not None:
                code += f"\tserving=serving_dataset_{index},\n"
            if resource.raw_data["training_dataset"] is not None:
                code += f"\ttraining=training_dataset_{index},\n"
            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKMonitor):
            parsed_config = parse_monitor_configuration(resource.raw_data["configuration"])
            detection_method = infer_detection_method(parsed_config)
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Monitor(\n"""
            code += f'\t"{resource.name}",\n'
            code += f"\tmonitor_type=aporia.MonitorType.{resource.type.name},\n"
            code += f"\tdetection_method=aporia.DetectionMethod.{detection_method.name},\n"

            if parsed_config.metric.type == MetricType.EUCLIDEAN_DISTANCE:
                code += "\tis_embedding_monitor=True,\n"

            # Identification configuration
            if parsed_config.identification.raw_inputs is not None:
                raw_inputs = parsed_config.identification.raw_inputs
                raw_inputs_array = (
                    f"""[{', '.join(f'"{raw_input}"' for raw_input in raw_inputs)}]"""
                )
                code += f"\traw_inputs={raw_inputs_array},\n"
            if parsed_config.identification.features is not None:
                features = parsed_config.identification.features
                features_array = f"""[{', '.join(f'"{feature}"' for feature in features)}]"""
                code += f"\tfeatures={features_array},\n"
            if parsed_config.identification.predictions is not None:
                predictions = parsed_config.identification.predictions
                predictions_array = (
                    f"""[{', '.join(f'"{prediction}"' for prediction in predictions)}]"""
                )
                code += f"\tpredictions={predictions_array},\n"
            if parsed_config.identification.models.version is not None:
                version_id = parsed_config.identification.models.version
                if version_id in (version_id_to_object or {}):
                    code += f"\tversion={version_id_to_object[version_id]},\n"
                else:
                    code += f'\tversion_id="{version_id}",\n'

            if parsed_config.identification.segment.group is not None:
                segment_id = parsed_config.identification.segment.group
                if segment_id in (segment_id_to_object or {}):
                    code += f"\tsegment={segment_id_to_object[segment_id]},\n"
                else:
                    code += f'\tsegment_id="{segment_id}",\n'

            if parsed_config.identification.segment.value is not None:
                segment_value = parsed_config.identification.segment.value
                if isinstance(segment_value, str):
                    segment_value = f'"{segment_value}"'
                code += f"\tsegment_value={segment_value},\n"

            # Dataset configuration:
            code += f"\tfocal=aporia.FocalConfiguration(**{parsed_config.focal.to_dict()}),\n"
            if parsed_config.baseline is not None:
                code += f"\tbaseline=aporia.BaselineConfiguration(**{parsed_config.baseline.to_dict()}),\n"

            # Metric Configuration
            if resource.type in [
                MonitorType.PERFORMANCE_DEGRADATION,
                MonitorType.METRIC_CHANGE,
            ]:
                code += f"\tmetric=aporia.MetricType.{parsed_config.metric.type.name},\n"
            metric_config = parsed_config.metric
            if metric_config.id is not None:
                custom_metric_id = metric_config.id
                if custom_metric_id in (custom_metric_id_to_object or {}):
                    code += f"\tcustom_metric={custom_metric_id_to_object[custom_metric_id]},\n"
                else:
                    code += f'\tcustom_metric_id="{custom_metric_id}",\n'
            if metric_config.average is not None:
                code += f"\taverage_method=aporia.AverageMethod.{metric_config.average.name},\n"
            if metric_config.metricAtK is not None:
                code += f"\tk={metric_config.metricAtK},\n"
            if metric_config.quantile is not None:
                code += f"\tquantile={metric_config.quantile},\n"
            if metric_config.threshold is not None:
                code += f"\tprediction_threshold={metric_config.threshold},\n"
            if metric_config.metricPerClass is not None:
                prediction_class = metric_config.metricPerClass
                if isinstance(prediction_class, str):
                    prediction_class = f'"{prediction_class}"'
                code += f"\tprediction_class={prediction_class},\n"

            # Logic Evaluation Configuration
            # TODO: Support multiple logic evaluations, if needed
            logic_evaluation_configuration = parsed_config.logicEvaluations[0]
            added_percentage = False
            use_percentage = False
            use_staleness = False
            if logic_evaluation_configuration.name is LogicEvaluationType.RATIO:
                use_percentage = True
            if logic_evaluation_configuration.name is LogicEvaluationType.MODEL_STALENESS:
                use_staleness = True
            if logic_evaluation_configuration.min is not None:
                if use_percentage:
                    if not added_percentage:
                        code += f"\tpercentage={(1-logic_evaluation_configuration.min)*100},\n"
                        added_percentage = True
                else:
                    code += f"\tmin={logic_evaluation_configuration.min},\n"
            if logic_evaluation_configuration.max is not None:
                if use_percentage:
                    if not added_percentage:
                        code += f"\tpercentage={(logic_evaluation_configuration.max-1)*100},\n"
                        added_percentage = True
                elif use_staleness:
                    code += (
                        f'\tstaleness_period="{logic_evaluation_configuration.max.to_string()}",\n'
                    )
                else:
                    code += f"\tmax={logic_evaluation_configuration.max},\n"
            if logic_evaluation_configuration.sensitivity is not None:
                code += f"\tsensitivity={logic_evaluation_configuration.sensitivity},\n"
            if logic_evaluation_configuration.testOnlyIncrease is not None:
                code += (
                    f"\talert_on_increase_only={logic_evaluation_configuration.testOnlyIncrease},\n"
                )
            if logic_evaluation_configuration.thresholds is not None:
                code += f"\tthresholds=aporia.ThresholdConfiguration(**{logic_evaluation_configuration.thresholds.to_dict()}),\n"
            if logic_evaluation_configuration.distance is not None:
                code += f"\tdistance={logic_evaluation_configuration.distance},\n"
            if logic_evaluation_configuration.new_values_count_threshold is not None:
                code += f"\tnew_values_count_threshold={logic_evaluation_configuration.new_values_count_threshold},\n"
            if logic_evaluation_configuration.new_values_ratio_threshold is not None:
                code += f"\tnew_values_ratio_threshold={logic_evaluation_configuration.new_values_ratio_threshold},\n"

            # Alert configuration
            added_alert = False
            for action in parsed_config.actions:
                if action.type is ActionType.ALERT:
                    if not added_alert:
                        emails = []
                        messaging = {}
                        code += f"\tseverity=aporia.Severity.{action.severity.name},\n"
                        for notification in action.notification:
                            if notification["type"] == "EMAIL":
                                emails.extend(notification["emails"])
                            else:
                                integration_type = (
                                    f"{MessagingIntegrationType(notification['type']).value}"
                                )
                                if integration_type not in messaging.keys():
                                    messaging[integration_type] = []
                                messaging[integration_type].append(notification["integration_id"])
                        if len(emails) > 0:
                            code += f"\temails={emails},\n"
                        if len(messaging) > 0:
                            code += f"\tmessaging={messaging},\n"
                        added_alert = True
                        if action.alertGroupByEntity is not None:
                            code += f"\talertGroupByEntity={action.alertGroupByEntity},\n"
                        if action.alertGroupByTime is not None:
                            code += f"\talertGroupByTime={action.alertGroupByTime},\n"
                        if action.alertGroupTimeQuantity is not None:
                            code += f"\talertGroupTimeQuantity={action.alertGroupTimeQuantity},\n"
                        if action.alertGroupTimeUnit is not None:
                            code += f"\talertGroupTimeUnit={action.alertGroupTimeUnit},\n"

            # Precondition Configuration
            added_min = False
            added_min_focal_datapoints = False
            added_min_baseline_datapoints = False
            for precondition in parsed_config.preConditions or []:
                if precondition.name is PreConditionType.FOCAL_DATA_VALUE_IN_RANGE:
                    if not added_min:
                        code += f"\tmin={precondition.min},\n"
                        added_min = True
                if precondition.name is PreConditionType.MIN_FOCAL_DATA_POINTS:
                    if not added_min_focal_datapoints:
                        code += f"\tmin_focal_prediction_count={precondition.value},\n"
                        added_min_focal_datapoints = True
                if precondition.name is PreConditionType.MIN_BASELINE_DATA_POINTS:
                    if not added_min_baseline_datapoints:
                        code += f"\tmin_baseline_prediction_count={precondition.value},\n"
                        added_min_baseline_datapoints = True

            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKModel):
            return (
                f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Model(
    "{resource.name}",
    type=aporia.ModelType.{resource.type.name},
    icon=aporia.ModelIcon.{resource.icon.name},
    color={f'aporia.ModelColor.{resource.color.name}' if isinstance(resource.color, ModelColor) else f'"{resource.color}"'},
    versions=[],
    segments=[],
    custom_metrics=[],
    monitors=[],
)""",
                self.__resource_name_to_python_name(resource.name),
            )
        raise TypeError(f"Unsupported type: {type(resource).__name__})")

    def __create_model_code(
        self,
        model: SDKModel,
        versions: List[Tuple[SDKVersion, List[SDKDataset]]],
        segments: List[SDKSegment],
        custom_metrics: List[SDKCustomMetric],
        monitors: List[SDKMonitor],
    ) -> Tuple[str, str]:
        model_code, model_name = self.__create_python_representation(model, 0)
        model_versions = []
        model_segments = []
        model_custom_metrics = []
        model_monitors = []
        versions_code = []
        segments_code = []
        custom_metrics_code = []
        monitors_code = []
        version_id_to_object = {}
        segment_id_to_object = {}
        custom_metric_id_to_object = {}
        for index, (version, datasets) in enumerate(versions):
            datasets_code = [
                self.__create_python_representation(dataset, index=index)[0] for dataset in datasets
            ]
            version_code, version_name = self.__create_python_representation(version, index=index)
            version_id_to_object[version.id] = version_name
            full_version_code = "\n".join([*datasets_code, version_code])
            versions_code.append(full_version_code)
            model_versions.append(version_name)
        for index, segment in enumerate(segments):
            segment_code, segment_name = self.__create_python_representation(segment, index=index)
            segment_id_to_object[segment.id] = segment_name
            segments_code.append(segment_code)
            model_segments.append(segment_name)
        for index, custom_metric in enumerate(custom_metrics):
            custom_metric_code, custom_metric_name = self.__create_python_representation(
                custom_metric, index=index
            )
            custom_metric_id_to_object[custom_metric.id] = custom_metric_name
            custom_metrics_code.append(custom_metric_code)
            model_custom_metrics.append(custom_metric_name)
        for index, monitor in enumerate(monitors):
            monitor_code, monitor_name = self.__create_python_representation(
                monitor,
                index=index,
                version_id_to_object=version_id_to_object,
                segment_id_to_object=segment_id_to_object,
                custom_metric_id_to_object=custom_metric_id_to_object,
            )
            monitors_code.append(monitor_code)
            model_monitors.append(monitor_name)

        model_code = model_code.replace("versions=[]", f"versions=[{', '.join(model_versions)}]")
        model_code = model_code.replace("segments=[]", f"segments=[{', '.join(model_segments)}]")
        model_code = model_code.replace(
            "custom_metrics=[]", f"custom_metrics=[{', '.join(model_custom_metrics)}]"
        )
        model_code = model_code.replace("monitors=[]", f"monitors=[{', '.join(model_monitors)}]")

        # Hack to allow newlines in f-string expressions
        newline = "\n"
        full_code = """import aporia.as_code as aporia
"""
        full_code += f"""
# Creating Model {model.name}
"""
        if len(versions) > 0:
            full_code += f"""
## Define Versions
{newline.join(versions_code)}
"""
        if len(segments) > 0:
            full_code += f"""
## Define Segments
{newline.join(segments_code)}
"""

        if len(custom_metrics) > 0:
            full_code += f"""
## Define Custom Metrics
{newline.join(custom_metrics_code)}
"""

        if len(monitors) > 0:
            full_code += f"""
## Define Monitors
{newline.join(monitors_code)}
"""

        full_code += "## Creating model\n"
        full_code += model_code

        full_code += f"""

def add_model(stack: aporia.Stack):
    stack.add({model_name})
"""

        return full_code, model_name

    def __create_resource_entry(
        self, resource: SDKBaseAporiaResource, parent_unique_id: Optional[str] = None
    ) -> Tuple[str, Dict]:
        data = deepcopy(resource.raw_data)
        data["_resource_type"] = type(resource).__name__
        if isinstance(resource, SDKDataset):
            resource_name = resource.raw_data["stage"]
        else:
            resource_name = resource.name
        name = f"{parent_unique_id + '/' if parent_unique_id is not None else ''}{resource_name.replace('/', '-')}"
        return (name, data)

    def __iterate_model(
        self, model: SDKModel, index: int
    ) -> Tuple[Dict[str, Dict[str, Any]], str, str]:
        resources = OrderedDict()
        model_name, data = self.__create_resource_entry(resource=model)
        resources[model_name] = data

        model_versions = []
        for version in model.get_versions():
            datasets = version.get_datasets()
            model_versions.append((version, datasets))
        model_segments = model.get_segments()
        model_custom_metrics = model.get_custom_metrics()
        model_monitors = model.get_monitors()

        # Add versions
        for version, datasets in model_versions:
            version_name, data = self.__create_resource_entry(
                resource=version, parent_unique_id=model_name
            )
            resources[version_name] = data
            # Add datasets
            for dataset in datasets:
                dataset_name, data = self.__create_resource_entry(
                    resource=dataset, parent_unique_id=version_name
                )
                resources[dataset_name] = data

        # Add segments
        for segment in model_segments:
            segment_name, data = self.__create_resource_entry(
                resource=segment, parent_unique_id=model_name
            )
            resources[segment_name] = data

        # Add custom metrics
        for custom_metric in model_custom_metrics:
            custom_metric_name, data = self.__create_resource_entry(
                resource=custom_metric, parent_unique_id=model_name
            )
            resources[custom_metric_name] = data

        # Add custom metrics
        for monitor in model_monitors:
            monitor_name, data = self.__create_resource_entry(
                resource=monitor, parent_unique_id=model_name
            )
            resources[monitor_name] = data

        model_code, model_name = self.__create_model_code(
            model=model,
            versions=model_versions,
            segments=model_segments,
            custom_metrics=model_custom_metrics,
            monitors=model_monitors,
        )

        return resources, model_code, model_name

    def _import(self, output_directory: str, model_names: Optional[List[str]]):
        # data_sources = self._aporia_api.get_data_sources()
        models = self._aporia_api.get_models()
        resources = OrderedDict()
        output_path = pathlib.Path(output_directory)
        output_path.mkdir(parents=True, exist_ok=True)

        models_code = {}

        for index, model in enumerate(models):
            if model_names is not None and model.name not in model_names:
                continue
            model_resources, model_code, model_name = self.__iterate_model(model=model, index=index)
            for name, data in model_resources.items():
                if name in resources.keys():
                    raise NameError(f"Resource {name} found duplicate!")
                resources[name] = data
            models_code[model_name] = model_code

        code = [
            "import os",
            "import aporia.as_code as aporia\n",
            "",
            'account = os.environ["APORIA_ACCOUNT"]',
            'workspace = os.environ["APORIA_WORKSPACE"]',
            'token = os.environ["APORIA_TOKEN"]',
            "",
            "stack = aporia.Stack(",
            "    account=account,",
            "    workspace=workspace,",
            "    token=token,",
            f'    base_url="{self._base_url}"',
            ")",
            "",
        ]

        code.append("\n# Registering Models")
        for model_name, model_code in models_code.items():
            with open(output_path / f"{model_name}.py", "w") as model_file:
                model_file.write(model_code)
                code.append(f"import {model_name}")
                code.append(f"{model_name}.add_model(stack)")

        code.append("\n# Applying Stack")
        code.append('stack.apply(config_path="stack.json", action="create", rollback=False)\n')

        with open(output_path / "stack.json", "w") as json_file:
            json_file.write(json.dumps(resources))

        with open(output_path / "main.py", "w") as python_file:
            python_file.write("\n".join(code))

    def _refresh_config(self):
        new_config = {}
        for unique_id, resource in self._output.items():
            resource_type = resource["_resource_type"]
            resource_name = f"{resource_type}/{unique_id.split('/')[-1]}"
            print("Refreshing resource: {:80.80s}".format(resource_name), end="\r")
            try:
                new_config[unique_id] = RESOURCE_TYPE_TO_RESOURCE_CLASS[resource_type].read(
                    client=self._client, id=resource["id"]
                )
                new_config[unique_id]["_resource_type"] = resource_type
            except EntityNotFoundException:
                print(f"{resource_type} - {unique_id} not found, removing from config")
            except Exception as e:
                print(f"issue refreshing data for: {resource_type} - {unique_id}")
                raise e
        print("Refresh complete" + (100 * " "))
        self._output = new_config

    def apply(
        self,
        action: StackAction = StackAction.CREATE,
        skip_preview: bool = False,
        yes: bool = False,
        rollback=False,  # BUG: Deletes all resources always
        yes_rollback=False,
        config_path: Optional[str] = None,
        debug: Optional[bool] = False,
        diff_format: str = "print",
        should_raise: bool = False,
    ):
        last_error = None
        action = StackAction(action)
        # TODO: Read config_path, and apply diffs. Then write to config_path
        if config_path is not None:
            try:
                with open(config_path, "r") as f:
                    stack_data = json.loads(f.read())
                    if "version" in stack_data and type(stack_data["version"]) is str:
                        stack_version = stack_data["version"]
                        self._output = stack_data["resources"]
                    else:
                        stack_version = "0"
                        self._output = stack_data
                    if sdk_version < stack_version:
                        raise Exception(
                            "SDK version is older than config version, please update aporia SDK (pip install --upgrade aporia)"
                        )
                    self._diff()
            except FileNotFoundError:
                # It's OK if the file wasn't found. It's probably the first run
                pass

        if action is StackAction.CREATE:
            if not skip_preview:
                self._preview()
            if not yes:
                if input("Do create? [Y/N] ").lower().strip() not in ["y"]:
                    print("Cancelling...")
                    return
            self._output, error = self._create(debug=debug)

            if error is not None:
                if debug:
                    traceback.print_exception(error, tb=error.__traceback__, value=error)
                else:
                    print(error)
                last_error = error
                if rollback:
                    if not yes_rollback:
                        if input("Delete created resources? [Y/N] ").lower().strip() in ["y"]:
                            print("Deleting...")
                            error = self._destroy()
                            if error is not None:
                                if debug:
                                    traceback.print_exception(
                                        error, tb=error.__traceback__, value=error
                                    )
                                else:
                                    print(error)
                                last_error = error
        elif action is StackAction.PREVIEW:
            self._preview()
            return
        elif action is StackAction.DELETE:
            if not skip_preview:
                self._preview(delete=True)
            # TODO: Change the way you use ResourceAction to support deletes, edits etc all at once, and previews for all
            if not yes:
                if input("Do delete? [Y/N] ").lower().strip() not in ["y"]:
                    print("Cancelling...")
                    return
            error = self._destroy()
            if error is not None:
                if debug:
                    traceback.print_exception(error, tb=error.__traceback__, value=error)
                else:
                    print(error)
                last_error = error
        elif action is StackAction.DIFF:
            self._get_diff_config(output_format=diff_format)
            return
        elif action is StackAction.REFRESH:
            self._refresh_config()
        else:
            raise ValueError("Unknown action")

        if config_path is not None:
            with open(config_path, "w") as f:
                f.write(json.dumps({"version": sdk_version, "resources": self._output}))

        if last_error is not None and should_raise:
            raise last_error

    def get_resource_id(self, unique_id: str) -> str:
        return self._output[unique_id]["id"]



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import FieldGroup
from aporia.sdk.metrics import MetricType
from aporia.sdk.monitors import AverageMethod
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import (
    BaselineConfiguration,
    BaseWidget,
    DisplayOptions,
    MetricConfiguration,
    MetricOverrideConfiguration,
    MetricParameters,
    VersionSelection,
    WidgetType,
)


class TimeSeriesWidgetDataOptionsAnnotations(BaseModel):
    phase: Literal[DatasetType.SERVING] = DatasetType.SERVING
    metric: Union[MetricType, str]
    version: Union[VersionSelection, List[VersionSelection]] = VersionSelection()
    field: Optional[Dict] = None
    fieldCategory: Optional[FieldGroup] = None
    granularity: Optional[
        str
    ] = None  # Optional[TimePeriod] = None # TODO: Find a way to implicitly convert a string to TimePeriod
    baseline: Optional[BaselineConfiguration] = None
    threshold: Optional[float] = None
    metricAtK: Optional[int] = None
    metricPerClass: Optional[str] = None
    average: Optional[AverageMethod] = None
    parameters: Optional[MetricParameters] = None
    isModelMetric: bool = True
    isCustomMetric: bool = False
    isAggregableMetric: bool = True
    isCodeBasedMetric: bool = False
    dataSegment: Optional[Dict] = None
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[
        Union[List[Union[int, float, str]], Union[int, float, str]]
    ] = None  # For dataSegmentValue and version, just one of them is a list
    dataSegmentBucketName: Optional[
        str
    ] = None  # This appears only when dataSegmentValue is not a list
    quantile: Optional[float] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class TimeSeriesWidgetDataOptionsFilters(BaseModel):
    annotations: TimeSeriesWidgetDataOptionsAnnotations


class TimeSeriesWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: TimeSeriesWidgetDataOptionsFilters


class TimeSeriesWidget(BaseWidget):
    type: Literal[WidgetType.TIME_SERIES] = WidgetType.TIME_SERIES
    dataOptions: List[TimeSeriesWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        metric: MetricConfiguration,
        # Mutually Exclusive
        version: Optional[Version] = None,
        versions: Optional[List[Optional[Version]]] = None,
        segment: Optional[Segment] = None,
        # Mutually Exclusive
        segment_value: Optional[Any] = None,
        segment_values: Optional[List[Any]] = None,
        granularity: Optional[str] = None,
    ) -> "TimeSeriesWidget":
        if version is not None and versions is not None:
            raise ValueError("Can't use both version and versions arguments")

        if segment_value is not None and segment_values is not None:
            raise ValueError("Can't use both segment_value and segment_values arguments")

        if segment is not None and segment_value is None and segment_values is None:
            if versions:
                raise ValueError(
                    "When using multiple versions and a segment, must specify segment_value"
                )
            segment_values = segment.values

        if segment_values is not None and versions is not None:
            raise ValueError("Can't pass segment_values when using multiple versions")

        if versions is None and segment is not None and segment_value is not None:
            segment_values = [segment_value]
            segment_value = None

        if (segment is None) != (segment_value is None and segment_values is None):
            raise ValueError(
                "For using segments, both segment and segment_value/segment_values arguments are needed"
            )

        if segment_values is not None:
            fixed_values = []
            for value in segment_values:
                if value in segment.get_widget_values():
                    fixed_values.append(value)
            if len(fixed_values) == 0:
                raise RuntimeError("No valid values given to segment")
            segment_values = fixed_values

        if version is None and versions is None:
            version_selection = VersionSelection()
        elif version is not None:
            version_selection = VersionSelection(id=version.id)
        else:
            version_selection = [
                VersionSelection(id=version.id) if version is not None else VersionSelection()
                for version in versions
            ]

        return TimeSeriesWidget(
            _internal_state={"metric": metric},
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                TimeSeriesWidgetDataOptions(
                    filters=TimeSeriesWidgetDataOptionsFilters(
                        annotations=TimeSeriesWidgetDataOptionsAnnotations(
                            metric=metric._get_metric(),
                            version=version_selection,
                            field=metric._get_field(),
                            fieldCategory=metric._get_field_category(),
                            baseline=metric._get_baseline(),
                            threshold=metric._get_threshold(),
                            metricAtK=metric._get_k(),
                            metricPerClass=metric._get_class(),
                            average=metric._get_average(),
                            parameters=metric._get_parameters(),
                            isModelMetric=metric._is_model_metric(),
                            isCustomMetric=metric._is_custom_metric(),
                            isAggregableMetric=metric._is_aggregable_metric(),
                            isCodeBasedMetric=metric._is_code_based_metric(),
                            dataSegment=segment.to_widget() if segment is not None else None,
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=(
                                (segment_value or segment_values) if segment is not None else None
                            ),
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] == segment_value
                                ][0]
                                if segment is not None and segment_value is not None
                                else None
                            ),
                            granularity=granularity,
                            quantile=metric._get_quantile(),
                        ),
                    )
                )
            ],
        )

    def compare(
        self,
        # Mutually Exclusive
        version: Optional[Version] = None,
        versions: Optional[List[Optional[Version]]] = None,
        metric_overrides: MetricOverrideConfiguration = MetricOverrideConfiguration(),
        segment: Optional[Segment] = None,
        # Mutually Exclusive
        segment_value: Optional[Any] = None,
        segment_values: Optional[List[Any]] = None,
    ) -> "TimeSeriesWidget":
        # TODO: Consider enforcing compare limits here
        original_metric: MetricConfiguration = self._internal_state["metric"]
        metric = metric_overrides._apply(original_metric)
        granularity = self.dataOptions[0].filters.annotations.granularity
        color = self.dataOptions[-1].display.color + 1
        if version is not None and versions is not None:
            raise ValueError("Can't use both version and versions arguments")

        if segment_value is not None and segment_values is not None:
            raise ValueError("Can't use both segment_value and segment_values arguments")

        if segment is not None and segment_value is None and segment_values is None:
            if versions:
                raise ValueError(
                    "When using multiple versions and a segment, must specify segment_value"
                )
            segment_values = segment.values

        if segment_values is not None and versions is not None:
            raise ValueError("Can't pass segment_values when using multiple versions")

        if versions is None and segment is not None and segment_value is not None:
            segment_values = [segment_value]
            segment_value = None

        if (segment is None) != (segment_value is None and segment_values is None):
            raise ValueError(
                "For using segments, both segment and segment_value/segment_values arguments are needed"
            )

        if segment_values is not None:
            fixed_values = []
            for value in segment_values:
                if value in segment.get_widget_values():
                    fixed_values.append(value)
            if len(fixed_values) == 0:
                raise RuntimeError("No valid values given to segment")
            segment_values = fixed_values

        if version is None and versions is None:
            version_selection = VersionSelection()
        elif version is not None:
            version_selection = VersionSelection(id=version.id)
        else:
            version_selection = [
                VersionSelection(id=version.id) if version is not None else VersionSelection()
                for version in versions
            ]

        self.dataOptions.append(
            TimeSeriesWidgetDataOptions(
                display=DisplayOptions(color=color),
                filters=TimeSeriesWidgetDataOptionsFilters(
                    annotations=TimeSeriesWidgetDataOptionsAnnotations(
                        metric=metric._get_metric(),
                        version=version_selection,
                        field=metric._get_field(),
                        fieldCategory=metric._get_field_category(),
                        baseline=metric._get_baseline(),
                        threshold=metric._get_threshold(),
                        metricAtK=metric._get_k(),
                        metricPerClass=metric._get_class(),
                        average=metric._get_average(),
                        parameters=metric._get_parameters(),
                        isModelMetric=metric._is_model_metric(),
                        isCustomMetric=metric._is_custom_metric(),
                        isAggregableMetric=metric._is_aggregable_metric(),
                        isCodeBasedMetric=metric._is_code_based_metric(),
                        dataSegment=segment.to_widget() if segment is not None else None,
                        dataSegmentId=segment.id if segment is not None else None,
                        dataSegmentName=segment.name if segment is not None else None,
                        dataSegmentValue=(
                            (segment_value or segment_values) if segment is not None else None
                        ),
                        dataSegmentBucketName=(
                            [
                                bucket["name"]
                                for bucket in segment.to_widget()["buckets"]
                                if bucket["value"] == segment_value
                            ][0]
                            if segment is not None and segment_value is not None
                            else None
                        ),
                        granularity=granularity,
                    ),
                ),
            )
        )
        return self



from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import (
    BaseWidget,
    DisplayOptions,
    FixedTimeframe,
    VersionSelection,
    WidgetType,
)


class SortType(str, Enum):
    DRIFT_SCORE = "drift_score"
    MISSING_VALUES = "missing_ratio"


class SortDirection(str, Enum):
    DESCENDING = "desc"
    ASCENDING = "asc"


class DataHealthTableWidgetDataOptionsAnnotations(BaseModel):
    phase: DatasetType
    version: VersionSelection = VersionSelection()
    sortBy: SortType = SortType.DRIFT_SCORE
    sortDirection: SortDirection = SortDirection.DESCENDING
    dataSegment: Optional[Dict] = None
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[Any] = None  # Optional[Union[int, float, str]] = None
    dataSegmentBucketName: Optional[str] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class DataHealthTableWidgetDataOptionsFilters(BaseModel):
    annotations: DataHealthTableWidgetDataOptionsAnnotations
    timeframe: Optional[Union[str, FixedTimeframe]] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class DataHealthTableWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: DataHealthTableWidgetDataOptionsFilters


class DataHealthTableWidget(BaseWidget):
    type: Literal[WidgetType.ANOMALY_TABLE] = WidgetType.ANOMALY_TABLE
    dataOptions: List[DataHealthTableWidgetDataOptions]  # Length must be 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        version: Optional[Version] = None,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        phase: DatasetType = DatasetType.SERVING,
        sort_by: SortType = SortType.DRIFT_SCORE,
        sort_direction: SortDirection = SortDirection.DESCENDING,
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
    ) -> "DataHealthTableWidget":
        if (segment is None) != (segment_value is None):
            raise ValueError(
                "For using segments, both segment and segment_value arguments are needed"
            )

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        return DataHealthTableWidget(
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                DataHealthTableWidgetDataOptions(
                    filters=DataHealthTableWidgetDataOptionsFilters(
                        timeframe=timeframe,
                        annotations=DataHealthTableWidgetDataOptionsAnnotations(
                            phase=phase,
                            version=version_selection,
                            sortBy=sort_by,
                            sortDirection=sort_direction,
                            dataSegment=segment.to_widget() if segment is not None else None,
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=segment_value if segment is not None else None,
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] == segment_value
                                ][0]
                                if segment is not None
                                else None
                            ),
                        ),
                    )
                )
            ],
        )

    def compare(
        self,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        phase: DatasetType = DatasetType.SERVING,
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
    ) -> "DataHealthTableWidget":
        color = self.dataOptions[-1].display.color + 1
        version = self.dataOptions[0].filters.annotations.version
        sort_by = self.dataOptions[0].filters.annotations.sortBy
        sort_direction = self.dataOptions[0].filters.annotations.sortDirection

        self.dataOptions.append(
            DataHealthTableWidgetDataOptions(
                display=DisplayOptions(color=color),
                filters=DataHealthTableWidgetDataOptionsFilters(
                    timeframe=timeframe,
                    annotations=DataHealthTableWidgetDataOptionsAnnotations(
                        phase=phase,
                        version=version,
                        sortBy=sort_by,
                        sortDirection=sort_direction,
                        dataSegment=segment.to_widget() if segment is not None else None,
                        dataSegmentId=segment.id if segment is not None else None,
                        dataSegmentName=segment.name if segment is not None else None,
                        dataSegmentValue=segment_value if segment is not None else None,
                        dataSegmentBucketName=(
                            [
                                bucket["name"]
                                for bucket in segment.to_widget()["buckets"]
                                if bucket["value"] == segment_value
                            ][0]
                            if segment is not None
                            else None
                        ),
                    ),
                ),
            ),
        )
        return self



import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel
from requests import Session

from aporia.sdk.client import Client
from aporia.sdk.datasets import DatasetType


class MetricType(str, Enum):
    COUNT = "count"
    MEAN = "mean"
    MEAN_LENGTH = "mean_length"
    MIN = "min"
    MEDIAN = "median"
    MIN_LENGTH = "min_length"
    MAX = "max"
    MAX_LENGTH = "max_length"
    SUM = "sum"
    MISSING_COUNT = "missing_count"
    MISSING_RATIO = "missing_ratio"
    TP_COUNT = "tp_count"
    TN_COUNT = "tn_count"
    FP_COUNT = "fp_count"
    FN_COUNT = "fn_count"
    ACCURACY = "accuracy"
    RECALL = "recall"
    PRECISION = "precision"
    AUC_ROC = "auc_roc"
    F1 = "f1"
    MSE = "mse"
    RMSE = "rmse"
    MAE = "mae"
    ACCURACY_AT_K = "accuracy_at_k"
    PRECISION_AT_K = "precision_at_k"
    RECALL_AT_K = "recall_at_k"
    MAP_AT_K = "map_at_k"
    MRR_AT_K = "mrr_at_k"
    NDCG_AT_K = "ndcg_at_k"
    JS_DISTANCE = "js_distance"
    KS_DISTANCE = "ks_distance"
    HELLINGER_DISTANCE = "hellinger_distance"
    EUCLIDEAN_DISTANCE = "euclidean_distance"
    PSI = "psi"
    ACCURACY_PER_CLASS = "accuracy_per_class"
    PRECISION_PER_CLASS = "precision_per_class"
    RECALL_PER_CLASS = "recall_per_class"
    F1_PER_CLASS = "f1_per_class"
    TP_COUNT_PER_CLASS = "tp_count_per_class"
    TN_COUNT_PER_CLASS = "tn_count_per_class"
    FP_COUNT_PER_CLASS = "fp_count_per_class"
    FN_COUNT_PER_CLASS = "fn_count_per_class"
    QUANTILE = "quantile"
    MAPE="mape"
    # Non-aggregable metrics
    AUUC = "auuc"
    # None-Dashboard metrics
    LAST_VERSION_CREATION = "last_version_creation"
    HISTOGRAM = "histogram"
    VALUE_COUNT = "value_count"
    MIN_MAX = "min_max"
    UNIQUE_VALUES = "unique_values"
    MISSING = "missing"
    # Custom metrics
    CUSTOM_METRIC = "custom_metric"
    CODE = "code"


class AverageMethod(str, Enum):
    MICRO = "micro"
    MACRO = "macro"
    WEIGHTED = "weighted"


def is_model_metric(metric: MetricType) -> bool:
    return metric is MetricType.COUNT


def is_custom_metric(metric: MetricType) -> bool:
    return metric is MetricType.CUSTOM_METRIC


def is_code_based_metric(metric: MetricType) -> bool:
    return metric is MetricType.CODE


def is_aggregable_metric(metric: MetricType) -> bool:
    return not (metric is MetricType.CODE or metric is MetricType.AUUC)


class TimeRange(BaseModel):
    start: datetime.datetime
    end: datetime.datetime  # "2023-05-12T00:00:00.000000+00:00"

    def serialize(self):
        return {
            "start": self.start.isoformat(timespec="microseconds") + "+00:00",
            "end": self.end.isoformat(timespec="microseconds") + "+00:00",
        }


class MetricSegment(BaseModel):
    id: str
    value: str  # TODO: Test with numeric auto segments

    def serialize(self) -> Dict:
        return {"id": self.id, "value": self.value}


class MetricDataset(BaseModel):
    dataset_type: DatasetType = DatasetType.SERVING
    time_range: Optional[TimeRange] = None
    model_version: Optional[str] = None
    segment: Optional[MetricSegment] = None

    def serialize(self) -> Dict:
        return {
            "dataset_type": self.dataset_type.value,
            "time_range": self.time_range.serialize() if self.time_range is not None else None,
            "model_version": self.model_version,
            "segment": self.segment.serialize() if self.segment is not None else None,
        }


class MetricResponse(BaseModel):
    id: str
    value: Any
    error: Optional[Union[Dict, List, str]]
    segment: Optional[MetricSegment]


class QueryResponse(BaseModel):
    metrics: List[MetricResponse]
    time_range: Optional[TimeRange] = None


class MetricParameters(BaseModel):
    dataset: MetricDataset
    # Metric identifier
    name: str
    # Parameters
    column: Optional[str] = None
    k: Optional[int] = None
    threshold: Optional[float] = None
    custom_metric_id: Optional[str] = None
    baseline: Optional[MetricDataset] = None
    value: Optional[Any] = None


class MetricError(Exception):
    def __init__(self, error, parameters: MetricParameters):
        super().__init__(error)
        self.parameters = parameters


class AporiaMetrics:
    def __init__(
        self,
        token: str,
        account_name: str,
        base_url: str = "https://platform.aporia.com",
        workspace_name: str = "default-workspace",
        requests_session: Optional[Session] = None,
    ) -> Dict:
        self.client = Client(
            base_url=f"{base_url}/api/v1/{account_name}/{workspace_name}",
            token=token,
            requests_session=requests_session,
        )

    def query(
        self,
        # Dataset identifiers
        model_id: str,
        dataset: MetricDataset,
        # Metric identifier
        metric_name: str,
        baseline: Optional[MetricDataset] = None,
        # Parameters
        column: Optional[str] = None,
        k: Optional[int] = None,
        threshold: Optional[float] = None,
        custom_metric_id: Optional[str] = None,
    ):
        metric_parameters = {}
        if column is not None:
            metric_parameters["column"] = column
        if k is not None:
            metric_parameters["k"] = k
        if threshold is not None:
            metric_parameters["threshold"] = threshold
        if custom_metric_id is not None:
            metric_parameters["id"] = custom_metric_id

        metric_datasets = {"data": dataset.serialize()}
        if baseline is not None:
            metric_datasets["baseline"] = baseline.serialize()

        response = self.client.send_request(
            "/query",
            method="POST",
            data={
                "model_id": model_id,
                "metrics": [
                    {
                        "id": "0",
                        "metric": metric_name,
                        "parameters": metric_parameters,
                        "datasets": metric_datasets,
                    }
                ],
            },
            url_search_replace=("/api/v1/", "/v1/metrics-reducer/"),
        )
        self.client.assert_response(response)

        result = [QueryResponse(**entry) for entry in response.json()]

        if result[0].metrics[0].error is not None:
            raise Exception(f"Error occured: {result[0].metrics[0].error}")
        return result[0].metrics[0].value

    def query_batch(self, model_id: str, metrics: List[MetricParameters]) -> List:
        metric_requests = []
        for i, metric in enumerate(metrics):
            metric_parameters = {}
            if metric.column is not None:
                metric_parameters["column"] = metric.column
            if metric.k is not None:
                metric_parameters["k"] = metric.k
            if metric.threshold is not None:
                metric_parameters["threshold"] = metric.threshold
            if metric.quantile is not None:
                metric_parameters["quantile"] = metric.quantile
            if metric.custom_metric_id is not None:
                metric_parameters["id"] = metric.custom_metric_id
            if metric.value is not None:
                metric_parameters["value"] = metric.value

            metric_datasets = {"data": metric.dataset.serialize()}
            if metric.baseline is not None:
                metric_datasets["baseline"] = metric.baseline.serialize()
            metric_requests.append(
                {
                    "id": str(i),
                    "metric": metric.name,
                    "parameters": metric_parameters,
                    "datasets": metric_datasets,
                }
            )

        response = self.client.send_request(
            "/query",
            method="POST",
            data={
                "model_id": model_id,
                "metrics": metric_requests,
            },
            url_search_replace=("/api/v1/", "/v1/metrics-reducer/"),
        )
        self.client.assert_response(response)

        result = [QueryResponse(**entry) for entry in response.json()]

        # Restore order, because different timeframes may change order
        metric_results = {}
        for entry in result:
            for metric in entry.metrics:
                metric_index = int(metric.id)
                if metric.error is not None:
                    metric_results[metric_index] = MetricError(
                        error=metric.error, parameters=metrics[metric_index]
                    )
                else:
                    metric_results[metric_index] = metric.value

        return [metric_results[i] for i in range(len(metric_results))]



import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from aporia.sdk.client import Client


class FieldType(str, Enum):
    BOOLEAN = "boolean"
    NUMERIC = "numeric"
    DATETIME = "datetime"
    CATEGORICAL = "categorical"
    TEXT = "text"
    ARRAY = "array"
    EMBEDDING = "embedding"
    IMAGE_URL = "image_url"
    NUMERIC_ARRAY = "numeric_array"


class FieldGroup(str, Enum):
    RAW_INPUTS = "raw_inputs"
    FEATURES = "features"
    PREDICTIONS = "predictions"
    ACTUALS = "actuals"


class Field(BaseModel):
    id: str
    name: str
    model_id: str
    type: FieldType
    group: FieldGroup
    properties: Optional[Dict[str, Any]] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime

    def to_widget(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "modelId": self.model_id,
            "type": self.type.value,
            "category": self.group.value,
            "properties": self.properties,
            "displayName": self.name,
        }

    @classmethod
    def get_model_fields(cls, client: Client, model_id: str) -> List["Field"]:
        response = client.send_request(
            "/fields",
            "GET",
            params={"model_id": model_id},
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        client.assert_response(response)
        return [cls(**field) for field in response.json()]



from http import HTTPStatus
from typing import Dict, Optional, Tuple, Union

import requests


class EntityNotFoundException(Exception):
    pass


class BackendRuntimeError(Exception):
    status_code: int
    body: Union[str, dict]

    def __init__(self, message: str, status_code: int, body: Union[str, dict]):
        super().__init__(message)
        self.status_code = status_code
        self.body = body


class Client:
    def __init__(
        self,
        base_url: str,
        token: str,
        debug: bool = False,
        requests_session: Optional[requests.Session] = None,
    ):
        self.base_url = base_url
        self.token = token
        self.debug = debug
        if requests_session is None:
            self._session = requests.Session()
        else:
            self._session = requests_session
        # Not setting Authorization header in session in case it's shared between multiple clients

    def send_request(
        self,
        url: str,
        method: str,
        data: Optional[Dict] = None,
        params: Optional[Dict[str, str]] = None,
        url_override: Optional[str] = None,
        url_search_replace: Optional[Tuple[str, str]] = None,
    ):
        if url_override is not None:
            target_url = url_override
        else:
            target_url = self.base_url + url

        if url_search_replace is not None:
            target_url = target_url.replace(url_search_replace[0], url_search_replace[1])

        if self.debug:
            print(f"{method} {target_url}, ({data})", end=" -> ")

        response = self._session.request(
            method=method,
            url=target_url,
            json=data,
            params=params,  # TODO: Start using this all over
            headers={"Authorization": f"Bearer {self.token}"},
        )

        if self.debug:
            print(f"{response.status_code}")

        return response

    def assert_response(self, response: requests.Response):
        if response.status_code == HTTPStatus.NOT_FOUND.value:
            raise EntityNotFoundException()
        if not response.ok:
            try:
                body = response.json()
            except Exception:
                body = response.content
            if self.debug:
                print(f"Request failed with status {response.status_code}, body: {body}")
            raise BackendRuntimeError(
                f"Request failed with status {response.status_code}, body: {body}",
                status_code=response.status_code,
                body=body,
            )



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import FieldGroup
from aporia.sdk.metrics import MetricType
from aporia.sdk.monitors import AverageMethod
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import (
    BaselineConfiguration,
    BaseWidget,
    DisplayOptions,
    FixedTimeframe,
    MetricConfiguration,
    MetricOverrideConfiguration,
    MetricParameters,
    VersionSelection,
    WidgetType,
)


class MetricCorrelationWidgetDataOptionsAnnotations(BaseModel):
    phase: Literal[DatasetType.SERVING] = DatasetType.SERVING
    version: Union[VersionSelection, List[VersionSelection]] = VersionSelection()
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[
        Union[List[Union[int, float, str]], Union[int, float, str]]
    ] = None  # For dataSegmentValue and version, just one of them is a list
    dataSegmentBucketName: Optional[
        str
    ] = None  # This appears only when dataSegmentValue is not a list
    # X axis
    xAxisMetric: Union[
        MetricType, str
    ]  # If the metric is a string, it is a custom metric/code-based metric ID
    isXAxisModelMetric: bool = False
    isXAxisCustomMetric: bool = False
    isXAxisCodeBasedMetric: bool = False
    isXAxisAggregableMetric: bool = True
    xAxisField: Optional[Dict] = None
    xAxisFieldCategory: Optional[FieldGroup] = None
    xAxisAverage: Optional[AverageMethod] = None
    xAxisMetricPerClass: Optional[str] = None
    xAxisMetricBaseline: Optional[BaselineConfiguration] = None
    xAxisThreshold: Optional[float] = None
    xAxisMetricAtK: Optional[int] = None
    xAxisQuantile: Optional[float] = None
    xAxisParameters: Optional[MetricParameters] = None
    # Y axis
    yAxisMetric: Union[
        MetricType, str
    ]  # If the metric is a string, it is a custom metric/code-based metric ID
    isYAxisModelMetric: bool = False
    isYAxisCustomMetric: bool = False
    isYAxisCodeBasedMetric: bool = False
    isYAxisAggregableMetric: bool = True
    yAxisField: Optional[Dict] = None
    yAxisFieldCategory: Optional[FieldGroup] = None
    yAxisAverage: Optional[AverageMethod] = None
    yAxisMetricPerClass: Optional[str] = None
    yAxisMetricBaseline: Optional[BaselineConfiguration] = None
    yAxisThreshold: Optional[float] = None
    yAxisMetricAtK: Optional[int] = None
    yAxisQuantile: Optional[float] = None
    yAxisParameters: Optional[MetricParameters] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricCorrelationWidgetDataOptionsFilters(BaseModel):
    annotations: MetricCorrelationWidgetDataOptionsAnnotations
    timeframe: Optional[Union[str, FixedTimeframe]] = None


class MetricCorrelationWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: MetricCorrelationWidgetDataOptionsFilters


class MetricCorrelationWidget(BaseWidget):
    type: Literal[WidgetType.METRIC_CORRELATION] = WidgetType.METRIC_CORRELATION
    dataOptions: List[MetricCorrelationWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        x_axis_metric: MetricConfiguration,
        y_axis_metric: MetricConfiguration,
        phase: DatasetType = DatasetType.SERVING,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        # Mutually Exclusive
        version: Optional[Version] = None,
        versions: Optional[List[Optional[Version]]] = None,
        segment: Optional[Segment] = None,
        # Mutually Exclusive
        segment_value: Optional[Any] = None,
        segment_values: Optional[List[Any]] = None,
    ) -> "MetricCorrelationWidget":
        if version and versions:
            raise ValueError("Can't use both version and versions arguments")

        if segment_value is not None and segment_values is not None:
            raise ValueError("Can't use both segment_value and segment_values arguments")

        if segment is not None and segment_value is None and segment_values is None:
            if versions:
                raise ValueError(
                    "When using multiple versions and a segment, must specify segment_value"
                )
            segment_values = segment.values

        if segment_values is not None and versions is not None:
            raise ValueError("Can't pass segment_values when using multiple versions")

        if versions is None and segment is not None and segment_value is not None:
            segment_values = [segment_value]
            segment_value = None

        if (segment is None) != (segment_value is None and segment_values is None):
            raise ValueError(
                "For using segments, both segment and segment_value/segment_values arguments are needed"
            )

        if segment_values is not None:
            fixed_values = []
            for value in segment_values:
                if value in segment.get_widget_values():
                    fixed_values.append(value)
            if len(fixed_values) == 0:
                raise RuntimeError("No valid values given to segment")
            segment_values = fixed_values

        if version is None and versions is None:
            version_selection = VersionSelection()
        elif version is not None:
            version_selection = VersionSelection(id=version.id)
        else:
            version_selection = [
                VersionSelection(id=version.id) if version is not None else VersionSelection()
                for version in versions
            ]

        return MetricCorrelationWidget(
            _internal_state={"x_metric": x_axis_metric, "y_metric": y_axis_metric},
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                MetricCorrelationWidgetDataOptions(
                    filters=MetricCorrelationWidgetDataOptionsFilters(
                        timeframe=timeframe,
                        annotations=MetricCorrelationWidgetDataOptionsAnnotations(
                            phase=phase,
                            version=version_selection,
                            # X axis
                            xAxisMetric=x_axis_metric._get_metric(),
                            xAxisField=x_axis_metric._get_field(),
                            xAxisFieldCategory=x_axis_metric._get_field_category(),
                            xAxisMetricBaseline=x_axis_metric._get_baseline(),
                            xAxisThreshold=x_axis_metric._get_threshold(),
                            xAxisMetricAtK=x_axis_metric._get_k(),
                            xAxisQuantile=x_axis_metric._get_quantile(),
                            xAxisMetricPerClass=x_axis_metric._get_class(),
                            xAxisAverage=x_axis_metric._get_average(),
                            xAxisParameters=x_axis_metric._get_parameters(),
                            isXAxisAggregableMetric=x_axis_metric._is_aggregable_metric(),
                            isXAxisModelMetric=x_axis_metric._is_model_metric(),
                            isXAxisCustomMetric=x_axis_metric._is_custom_metric(),
                            isXAxisCodeBasedMetric=x_axis_metric._is_code_based_metric(),
                            # Y axis
                            yAxisMetric=y_axis_metric._get_metric(),
                            yAxisField=y_axis_metric._get_field(),
                            yAxisFieldCategory=y_axis_metric._get_field_category(),
                            yAxisMetricBaseline=y_axis_metric._get_baseline(),
                            yAxisThreshold=y_axis_metric._get_threshold(),
                            yAxisMetricAtK=y_axis_metric._get_k(),
                            yAxisQuantile=y_axis_metric._get_quantile(),
                            yAxisMetricPerClass=y_axis_metric._get_class(),
                            yAxisAverage=y_axis_metric._get_average(),
                            yAxisParameters=y_axis_metric._get_parameters(),
                            isYAxisAggregableMetric=y_axis_metric._is_aggregable_metric(),
                            isYAxisModelMetric=y_axis_metric._is_model_metric(),
                            isYAxisCustomMetric=y_axis_metric._is_custom_metric(),
                            isYAxisCodeBasedMetric=y_axis_metric._is_code_based_metric(),
                            # Segment
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=(
                                (segment_value or segment_values) if segment is not None else None
                            ),
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] == segment_value
                                ][0]
                                if segment is not None and segment_value is not None
                                else None
                            ),
                        ),
                    )
                )
            ],
        )

    def compare(
        self,
        phase: DatasetType = DatasetType.SERVING,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        x_axis_metric_overrides: MetricOverrideConfiguration = MetricOverrideConfiguration(),
        y_axis_metric_overrides: MetricOverrideConfiguration = MetricOverrideConfiguration(),
        # Mutually Exclusive
        version: Optional[Version] = None,
        versions: Optional[List[Optional[Version]]] = None,
        segment: Optional[Segment] = None,
        # Mutually Exclusive
        segment_value: Optional[Any] = None,
        segment_values: Optional[List[Any]] = None,
    ) -> "MetricCorrelationWidget":
        # TODO: Consider enforcing compare limits here
        original_x_axis_metric: MetricConfiguration = self._internal_state["x_metric"]
        x_axis_metric = x_axis_metric_overrides._apply(original_x_axis_metric)
        original_y_axis_metric: MetricConfiguration = self._internal_state["y_metric"]
        y_axis_metric = y_axis_metric_overrides._apply(original_y_axis_metric)
        color = self.dataOptions[-1].display.color + 1

        if version and versions:
            raise ValueError("Can't use both version and versions arguments")

        if segment_value is not None and segment_values is not None:
            raise ValueError("Can't use both segment_value and segment_values arguments")

        if segment is not None and segment_value is None and segment_values is None:
            if versions:
                raise ValueError(
                    "When using multiple versions and a segment, must specify segment_value"
                )
            segment_values = segment.values

        if segment_values is not None and versions is not None:
            raise ValueError("Can't pass segment_values when using multiple versions")

        if versions is None and segment is not None and segment_value is not None:
            segment_values = [segment_value]
            segment_value = None

        if (segment is None) != (segment_value is None and segment_values is None):
            raise ValueError(
                "For using segments, both segment and segment_value/segment_values arguments are needed"
            )

        if segment_values is not None:
            fixed_values = []
            for value in segment_values:
                if value in segment.get_widget_values():
                    fixed_values.append(value)
            if len(fixed_values) == 0:
                raise RuntimeError("No valid values given to segment")
            segment_values = fixed_values

        if version is None and versions is None:
            version_selection = VersionSelection()
        elif version is not None:
            version_selection = VersionSelection(id=version.id)
        else:
            version_selection = [
                VersionSelection(id=version.id) if version is not None else VersionSelection()
                for version in versions
            ]

        self.dataOptions.append(
            MetricCorrelationWidgetDataOptions(
                display=DisplayOptions(color=color),
                filters=MetricCorrelationWidgetDataOptionsFilters(
                    timeframe=timeframe,
                    annotations=MetricCorrelationWidgetDataOptionsAnnotations(
                        phase=phase,
                        version=version_selection,
                        # X axis
                        xAxisMetric=x_axis_metric._get_metric(),
                        xAxisField=x_axis_metric._get_field(),
                        xAxisFieldCategory=x_axis_metric._get_field_category(),
                        xAxisMetricBaseline=x_axis_metric._get_baseline(),
                        xAxisThreshold=x_axis_metric._get_threshold(),
                        xAxisMetricAtK=x_axis_metric._get_k(),
                        xAxisQuantile=x_axis_metric._get_quantile(),
                        xAxisMetricPerClass=x_axis_metric._get_class(),
                        xAxisAverage=x_axis_metric._get_average(),
                        xAxisParameters=x_axis_metric._get_parameters(),
                        isXAxisAggregableMetric=x_axis_metric._is_aggregable_metric(),
                        isXAxisModelMetric=x_axis_metric._is_model_metric(),
                        isXAxisCustomMetric=x_axis_metric._is_custom_metric(),
                        isXAxisCodeBasedMetric=x_axis_metric._is_code_based_metric(),
                        # Y axis
                        yAxisMetric=y_axis_metric._get_metric(),
                        yAxisField=y_axis_metric._get_field(),
                        yAxisFieldCategory=y_axis_metric._get_field_category(),
                        yAxisMetricBaseline=y_axis_metric._get_baseline(),
                        yAxisThreshold=y_axis_metric._get_threshold(),
                        yAxisMetricAtK=y_axis_metric._get_k(),
                        yAxisQuantile=y_axis_metric._get_quantile(),
                        yAxisMetricPerClass=y_axis_metric._get_class(),
                        yAxisAverage=y_axis_metric._get_average(),
                        yAxisParameters=y_axis_metric._get_parameters(),
                        isYAxisAggregableMetric=y_axis_metric._is_aggregable_metric(),
                        isYAxisModelMetric=y_axis_metric._is_model_metric(),
                        isYAxisCustomMetric=y_axis_metric._is_custom_metric(),
                        isYAxisCodeBasedMetric=y_axis_metric._is_code_based_metric(),
                        # Segment
                        dataSegmentId=segment.id if segment is not None else None,
                        dataSegmentName=segment.name if segment is not None else None,
                        dataSegmentValue=(
                            (segment_value or segment_values) if segment is not None else None
                        ),
                        dataSegmentBucketName=(
                            [
                                bucket["name"]
                                for bucket in segment.to_widget()["buckets"]
                                if bucket["value"] == segment_value
                            ][0]
                            if segment is not None and segment_value is not None
                            else None
                        ),
                    ),
                ),
            )
        )
        return self



from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import Field as SDKField
from aporia.sdk.fields import FieldGroup
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import BaseWidget, DisplayOptions, VersionSelection, WidgetType


class HistogramOverTimeWidgetDataOptionsAnnotations(BaseModel):
    field: Dict
    phase: Literal[DatasetType.SERVING] = DatasetType.SERVING
    version: VersionSelection = VersionSelection()
    fieldCategory: FieldGroup
    granularity: Optional[str] = None
    dataSegment: Optional[Dict] = None
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[Any] = None  # Optional[Union[int, float, str]] = None
    dataSegmentBucketName: Optional[str] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class HistogramOverTimeWidgetDataOptionsFilters(BaseModel):
    annotations: HistogramOverTimeWidgetDataOptionsAnnotations


class HistogramOverTimeWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: HistogramOverTimeWidgetDataOptionsFilters


class HistogramOverTimeWidget(BaseWidget):
    type: Literal[WidgetType.TIME_SERIES_HISTOGRAM] = WidgetType.TIME_SERIES_HISTOGRAM
    dataOptions: List[HistogramOverTimeWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        field: SDKField,
        version: Optional[Version] = None,
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
        granularity: Optional[str] = None,
    ) -> "HistogramOverTimeWidget":
        if (segment is None) != (segment_value is None):
            raise ValueError(
                "For using segments, both segment and segment_value arguments are needed"
            )

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        return HistogramOverTimeWidget(
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                HistogramOverTimeWidgetDataOptions(
                    filters=HistogramOverTimeWidgetDataOptionsFilters(
                        annotations=HistogramOverTimeWidgetDataOptionsAnnotations(
                            version=version_selection,
                            field=field.to_widget(),
                            granularity=granularity,
                            fieldCategory=field.group,
                            dataSegment=segment.to_widget() if segment is not None else None,
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=segment_value if segment is not None else None,
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] == segment_value
                                ][0]
                                if segment is not None
                                else None
                            ),
                        ),
                    )
                )
            ],
        )



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import FieldGroup
from aporia.sdk.metrics import MetricType
from aporia.sdk.monitors import AverageMethod
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import (
    BaselineConfiguration,
    BaseWidget,
    DisplayOptions,
    FixedTimeframe,
    MetricConfiguration,
    MetricOverrideConfiguration,
    MetricParameters,
    VersionSelection,
    WidgetType,
)


class MetricBySegmentWidgetDataOptionsAnnotations(BaseModel):
    phase: Literal[DatasetType.SERVING] = DatasetType.SERVING
    metric: Union[
        MetricType, str
    ]  # If the metric is a string, it is a custom metric/code-based metric ID
    version: VersionSelection = VersionSelection()
    field: Optional[Dict] = None
    fieldCategory: Optional[FieldGroup] = None
    baseline: Optional[BaselineConfiguration] = None
    threshold: Optional[float] = None
    metricAtK: Optional[int] = None
    quantile: Optional[float] = None
    metricPerClass: Optional[str] = None
    average: Optional[AverageMethod] = None
    parameters: Optional[MetricParameters] = None
    isModelMetric: bool = True
    isCustomMetric: bool = False
    isAggregableMetric: bool = True
    isCodeBasedMetric: bool = False
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[Union[List[Union[int, float, str]], Union[int, float, str]]] = None
    dataSegmentBucketName: Optional[List[str]] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricBySegmentWidgetDataOptionsFilters(BaseModel):
    annotations: MetricBySegmentWidgetDataOptionsAnnotations
    timeframe: Optional[Union[str, FixedTimeframe]] = None


class MetricBySegmentWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: MetricBySegmentWidgetDataOptionsFilters


class MetricBySegmentWidget(BaseWidget):
    type: Literal[WidgetType.METRIC_BY_SEGMENT] = WidgetType.METRIC_BY_SEGMENT
    dataOptions: List[MetricBySegmentWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        metric: MetricConfiguration,
        segment: Segment,
        segment_values: List[Any],
        version: Optional[Version] = None,
        phase: DatasetType = DatasetType.SERVING,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
    ) -> "MetricBySegmentWidget":
        fixed_values = []
        for value in segment_values:
            if value in segment.get_widget_values():
                fixed_values.append(value)
        if len(fixed_values) == 0:
            raise RuntimeError("No valid values given to segment")
        segment_values = fixed_values

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        return MetricBySegmentWidget(
            _internal_state={"metric": metric},
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                MetricBySegmentWidgetDataOptions(
                    filters=MetricBySegmentWidgetDataOptionsFilters(
                        timeframe=timeframe,
                        annotations=MetricBySegmentWidgetDataOptionsAnnotations(
                            metric=metric._get_metric(),
                            version=version_selection,
                            phase=phase,
                            field=metric._get_field(),
                            fieldCategory=metric._get_field_category(),
                            baseline=metric._get_baseline(),
                            threshold=metric._get_threshold(),
                            metricAtK=metric._get_k(),
                            quantile=metric._get_quantile(),
                            metricPerClass=metric._get_class(),
                            average=metric._get_average(),
                            parameters=metric._get_parameters(),
                            isModelMetric=metric._is_model_metric(),
                            isCustomMetric=metric._is_custom_metric(),
                            isAggregableMetric=metric._is_aggregable_metric(),
                            isCodeBasedMetric=metric._is_code_based_metric(),
                            dataSegmentId=segment.id,
                            dataSegmentName=segment.name,
                            dataSegmentValue=segment_values,
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] in segment_values
                                ]
                            ),
                        ),
                    )
                )
            ],
        )

    def compare(
        self,
        version: Optional[Version] = None,
        metric: Optional[MetricConfiguration] = None,
        metric_overrides: Optional[MetricOverrideConfiguration] = None,
        phase: DatasetType = DatasetType.SERVING,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
    ) -> "MetricBySegmentWidget":
        # TODO: Consider enforcing compare limits here
        original_metric: MetricConfiguration = self._internal_state["metric"]
        if metric is not None and metric_overrides is not None:
            raise ValueError("Can't provide both metric and metric_overrides.")
        if metric is not None:
            metric = metric
        elif metric_overrides is not None:
            metric = metric_overrides._apply(original_metric)
        else:
            metric = MetricOverrideConfiguration()._apply(original_metric)
        color = self.dataOptions[-1].display.color + 1
        data_segment_id = self.dataOptions[-1].filters.annotations.dataSegmentId
        data_segment_name = self.dataOptions[-1].filters.annotations.dataSegmentName
        data_segment_value = self.dataOptions[-1].filters.annotations.dataSegmentValue
        data_segment_bucket_name = self.dataOptions[-1].filters.annotations.dataSegmentBucketName

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        self.dataOptions.append(
            MetricBySegmentWidgetDataOptions(
                display=DisplayOptions(color=color),
                filters=MetricBySegmentWidgetDataOptionsFilters(
                    timeframe=timeframe,
                    annotations=MetricBySegmentWidgetDataOptionsAnnotations(
                        metric=metric._get_metric(),
                        version=version_selection,
                        phase=phase,
                        field=metric._get_field(),
                        fieldCategory=metric._get_field_category(),
                        baseline=metric._get_baseline(),
                        threshold=metric._get_threshold(),
                        metricAtK=metric._get_k(),
                        quantile=metric._get_quantile(),
                        metricPerClass=metric._get_class(),
                        average=metric._get_average(),
                        parameters=metric._get_parameters(),
                        isModelMetric=metric._is_model_metric(),
                        isCustomMetric=metric._is_custom_metric(),
                        isAggregableMetric=metric._is_aggregable_metric(),
                        isCodeBasedMetric=metric._is_code_based_metric(),
                        dataSegmentId=data_segment_id,
                        dataSegmentName=data_segment_name,
                        dataSegmentValue=data_segment_value,
                        dataSegmentBucketName=data_segment_bucket_name,
                    ),
                ),
            )
        )
        return self



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, validator

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.widgets import BaseWidget
from aporia.sdk.widgets.base import FixedTimeframe, WidgetType
from aporia.sdk.widgets.data_health_table_widget import DataHealthTableWidget
from aporia.sdk.widgets.distribution_widget import DistributionWidget
from aporia.sdk.widgets.histogram_over_time_widget import HistogramOverTimeWidget
from aporia.sdk.widgets.metric_by_segment_widget import MetricBySegmentWidget
from aporia.sdk.widgets.metric_correlation_widget import MetricCorrelationWidget
from aporia.sdk.widgets.metric_widget import MetricWidget
from aporia.sdk.widgets.text_widget import TextWidget
from aporia.sdk.widgets.timeseries_widget import TimeSeriesWidget

ALLOWED_RELATIVE_TIMEFRAMES = ["1h", "4h", "1d", "2d", "1w", "2w", "1M", "2M", "3M"]


class SkipDaysUnit(str, Enum):
    DAYS = "Days"
    WEEKS = "Weeks"


class DashboardGlobalFilters(BaseModel):
    timeframe: Union[str, FixedTimeframe]
    version_id: Optional[str] = None
    data_segment_id: Optional[str] = None
    data_segment_value: Optional[Any] = None
    skip_days: Optional[int] = None
    skip_days_unit: Optional[SkipDaysUnit] = None

    @validator("timeframe")
    def _validate_timeframe(cls, value, values):
        if isinstance(value, str):
            if value not in ALLOWED_RELATIVE_TIMEFRAMES:
                raise ValueError(
                    f"Timeframe '{value}' is invalid. Available values are: {ALLOWED_RELATIVE_TIMEFRAMES}"
                )
        return value

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


Widget = Union[
    TextWidget,
    DataHealthTableWidget,
    DistributionWidget,
    HistogramOverTimeWidget,
    MetricWidget,
    TimeSeriesWidget,
    MetricBySegmentWidget,
    MetricCorrelationWidget,
]

WIDGET_TYPE_TO_CONFIG_CLASS = {
    WidgetType.TEXT: TextWidget,
    WidgetType.DISTRIBUTION: DistributionWidget,
    WidgetType.METRIC: MetricWidget,
    WidgetType.TIME_SERIES: TimeSeriesWidget,
    WidgetType.TIME_SERIES_HISTOGRAM: HistogramOverTimeWidget,
    WidgetType.ANOMALY_TABLE: DataHealthTableWidget,
    WidgetType.METRIC_BY_SEGMENT: MetricBySegmentWidget,
    WidgetType.METRIC_CORRELATION: MetricCorrelationWidget,
}


class DashboardConfiguration(BaseModel):
    widgets: List[Widget]
    global_filters: DashboardGlobalFilters = DashboardGlobalFilters(timeframe="1w")

    @validator("widgets", pre=True)
    @classmethod
    def _parse_connection(cls, value: Any, values: Dict[str, Any]) -> List[Widget]:
        return dashboard_widgets_pre_validator(value, values)


def dashboard_widgets_pre_validator(value: Any, values: Dict[str, Any]) -> List[Widget]:
    """Reads data of a dataset.

    Args:
        value: connection dict to validate
        values: other attributes of DataSource (type)

    Returns:
        List[Widget] object
    """
    # Parsing `widget` requires `type` to be parsed
    widgets = []

    if not isinstance(value, list):
        raise ValueError("expected a list containing widget configurations")

    for widget_configuration in value:
        if isinstance(widget_configuration, BaseWidget):
            widgets.append(widget_configuration)
            continue

        if "type" not in widget_configuration:
            raise ValueError("Received widget with no type")

        if widget_configuration["type"] not in WIDGET_TYPE_TO_CONFIG_CLASS:
            raise ValueError(f"Unsupported widget type {widget_configuration['type']}")

        if not isinstance(widget_configuration, dict):
            raise ValueError("expected a dict containing data source configuration")

        config_cls = WIDGET_TYPE_TO_CONFIG_CLASS[widget_configuration["type"]]
        widgets.append(config_cls(**widget_configuration))

    return widgets


class Dashboard(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.id = data["id"]
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.definition = DashboardConfiguration(**data["definition"])
        self.model_id: Optional[str] = data.get("model_id")
        self.raw_data = data

    @classmethod
    def get_all(cls, client: Client, model_id: Optional[str] = None) -> List["Dashboard"]:
        response = client.send_request(
            "/dashboards",
            "GET",
            params={"model_id": model_id} if model_id is not None else None,
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        description: str,
        model_id: Optional[str],
        definition: DashboardConfiguration,
        owner: Optional[str] = None,
    ) -> "Dashboard":
        response = client.send_request(
            "/dashboards",
            "POST",
            {
                "name": name,
                "description": description,
                "model_id": model_id,
                "definition": definition.dict(),
                "owner": owner,
            },
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Dashboard":
        response = client.send_request(
            f"/dashboards/{id}",
            "GET",
        )
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(self, **kwargs):
        args = {}

        if "name" in kwargs:
            args["name"] = kwargs["name"]

        if "description" in kwargs:
            args["description"] = kwargs["description"]

        if "owner" in kwargs:
            args["owner"] = kwargs["owner"]

        if "definition" in kwargs:
            args["definition"] = kwargs["definition"].dict()

        response = self.client.send_request(
            f"/dashboards/{self.id}",
            "PUT",
            args,
        )
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/dashboards/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/dashboards/{id}", "DELETE")
        client.assert_response(response)



import datetime
from enum import Enum
from typing import Dict, Optional, Tuple, Union
import uuid

from pydantic import BaseModel, Field, PrivateAttr, root_validator, validator

from aporia.sdk.custom_metrics import CustomMetric
from aporia.sdk.fields import Field as SDKField
from aporia.sdk.fields import FieldGroup, FieldType
from aporia.sdk.metrics import (
    AverageMethod,
    is_aggregable_metric,
    is_code_based_metric,
    is_custom_metric,
    is_model_metric,
    MetricType,
)


class FixedTimeframe(BaseModel):
    to: datetime.datetime
    from_impl: datetime.datetime = Field(alias="from")

    @validator("to", "from_impl")
    def _validate_utc(cls, value, values):
        if value.tzinfo != datetime.timezone.utc:
            raise ValueError("Timezones must be in UTC")
        return value

    def dict(self, *args, **kwargs):
        return {"to": self.to.isoformat(), "from": self.from_impl.isoformat()}


class DisplayOptions(BaseModel):
    color: int = 0  # 1 for compare


class VersionSelection(BaseModel):
    id: str = "*"


class BaselineType(str, Enum):
    TRAINING = "training"
    RELATIVE_TIME_PERIOD = "relativeTimePeriod"


class TimeUnit(str, Enum):
    DAY = "day"
    MONTH = "month"


class BaselineConfiguration(BaseModel):
    type: BaselineType
    unit: Optional[TimeUnit] = None
    duration: Optional[int] = None

    @validator("unit")
    def _validate_unit(cls, value, values) -> Optional[TimeUnit]:
        if values["type"] == "training" and value is not None:
            raise ValueError("Unit can't be used with training baseline")
        return value

    @validator("duration")
    def _validate_duration(cls, value, values) -> Optional[int]:
        if values["type"] == "training" and value is not None:
            raise ValueError("Unit can't be used with training baseline")
        return value

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricParameters(BaseModel):
    id: str


class WidgetType(str, Enum):
    TEXT = "text"
    ANOMALY_TABLE = "anomaly-table"
    DISTRIBUTION = "distribution"
    TIME_SERIES_HISTOGRAM = "time-series-histogram"
    METRIC = "metric"
    TIME_SERIES = "time-series"
    METRIC_BY_SEGMENT = "metric-by-segment"
    METRIC_CORRELATION = "metric-correlation"


class BaseWidget(BaseModel):
    i: str = Field(default_factory=lambda: str(uuid.uuid4()))
    h: int
    w: int  # Capped at 12
    x: int  # Capped at 11?
    y: int
    name: str
    type: WidgetType
    moved: bool = False
    resizable: bool = True
    dependencies: Optional[
        Dict
    ] = None  # TODO: Investigate this, and consider if you want to raise error if it has issues
    _internal_state: Dict = PrivateAttr(default_factory=dict)

    def __init__(self, **data):
        internal_state = data.pop("_internal_state", None)
        super().__init__(**data)
        if internal_state is not None:
            self._internal_state = internal_state

    @validator("x")
    def _validate_x(cls, value, values):
        if value < 0:
            raise ValueError("x must be non-negative")
        if value >= 12:
            raise ValueError("The grid's x-axis is limited at 12 cells")
        return value

    @validator("y")
    def _validate_y(cls, value, values):
        if value < 0:
            raise ValueError("y must be non-negative")
        return value

    @validator("w")
    def _validate_w(cls, value, values):
        if value <= 0:
            raise ValueError("w must be positive")
        if value > 12:
            raise ValueError("The grid's x-axis is limited at 12 cells")
        return value

    @validator("h")
    def _validate_h(cls, value, values):
        if value <= 0:
            raise ValueError("h must be positive")
        return value

    @classmethod
    def create(
        cls, position: Tuple[int, int], size: Tuple[int, int], *args, **kwargs
    ) -> "BaseWidget":
        raise NotImplementedError()

    def dict(self, *args, **kwargs):
        return {
            k: v
            for k, v in super().dict(*args, **kwargs).items()
            if v is not None and k != "_internal_state"
        }


class MetricBaseModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True  # For CustomMetric as an object


class MetricConfiguration(MetricBaseModel):
    metric: MetricType
    field: Optional[SDKField] = None
    custom_metric: Optional[CustomMetric] = None
    average: Optional[AverageMethod] = None
    threshold: Optional[float] = None
    k: Optional[int] = None
    class_name: Optional[str] = None
    baseline: Optional[BaselineConfiguration] = None
    quantile: Optional[float] = None

    def _get_metric(self) -> Union[MetricType, str]:
        if self.metric is MetricType.CUSTOM_METRIC:
            return self.custom_metric.id
        if self.metric is MetricType.CODE:
            raise NotImplementedError("Code-based metrics currently not supported")
        return self.metric

    def _get_field(self) -> Optional[Dict]:
        if self.field is not None:
            return self.field.to_widget()
        return None

    def _get_field_category(self) -> Optional[FieldGroup]:
        if self.field is not None:
            return self.field.group
        return None

    def _get_parameters(self) -> Optional[MetricParameters]:
        if self.custom_metric is not None:
            return MetricParameters(id=self.custom_metric.id)
        return None

    def _get_average(self) -> Optional[AverageMethod]:
        return self.average

    def _get_threshold(self) -> Optional[float]:
        return self.threshold

    def _get_k(self) -> Optional[int]:
        return self.k
    
    def _get_quantile(self) -> Optional[float]:
        return self.quantile

    def _get_class(self) -> Optional[str]:
        return self.class_name

    def _get_baseline(self) -> Optional[BaselineConfiguration]:
        return self.baseline

    def _is_model_metric(self) -> bool:
        return is_model_metric(self.metric)

    def _is_aggregable_metric(self) -> bool:
        return is_aggregable_metric(self.metric)

    def _is_custom_metric(self) -> bool:
        return is_custom_metric(self.metric)

    def _is_code_based_metric(self) -> bool:
        return is_code_based_metric(self.metric)

    @validator("threshold")
    def _validate_threshold(cls, value, values):
        if value is not None:
            if value < 0 or value > 1:
                raise ValueError("Threshold must be between 0 and 1, inclusive")
            value = round(value, 2)
        return value

    @validator("k")
    def _validate_k(cls, value, values):
        if value is not None:
            if not (1 <= value <= 12):
                raise ValueError("K must be between 1 and 12, inclusive")
        return value

    @root_validator()
    def _validate_metric_parameters(cls, values):
        value = values["metric"]
        if value is MetricType.CUSTOM_METRIC and values["custom_metric"] is None:
            raise ValueError("Must provide `custom_metric` parameter")

        if value is MetricType.CODE:
            # TODO: Add support when code-based metrics are added to the SDK as an object
            raise NotImplementedError(
                "Code-based metrics are currently not supported through the SDK"
            )

        if (
            value not in [MetricType.COUNT, MetricType.CUSTOM_METRIC, MetricType.CODE]
            and values["field"] is None
        ):
            raise ValueError("Must provide `field` parameter")

        if (
            value
            in [
                MetricType.ACCURACY_AT_K,
                MetricType.PRECISION_AT_K,
                MetricType.RECALL_AT_K,
                MetricType.MAP_AT_K,
                MetricType.MRR_AT_K,
                MetricType.NDCG_AT_K,
            ]
            and values["k"] is None
        ):
            raise ValueError("Must provide `k` parameter` for ranking metrics")

        if (
            value
            in [
                MetricType.JS_DISTANCE,
                MetricType.KS_DISTANCE,
                MetricType.HELLINGER_DISTANCE,
                MetricType.EUCLIDEAN_DISTANCE,
                MetricType.PSI,
            ]
            and values["baseline"] is None
        ):
            raise ValueError("Must provide `baseline` parameter for drift metrics")

        if (
            value
            in [
                MetricType.ACCURACY_PER_CLASS,
                MetricType.PRECISION_PER_CLASS,
                MetricType.RECALL_PER_CLASS,
                MetricType.F1_PER_CLASS,
                MetricType.TP_COUNT_PER_CLASS,
                MetricType.TN_COUNT_PER_CLASS,
                MetricType.FP_COUNT_PER_CLASS,
                MetricType.FN_COUNT_PER_CLASS,
            ]
            and values["class_name"] is None
        ):
            raise ValueError("Must provide `class_name` parameter for multiclass per-class metrics")

        if value in [
            MetricType.TP_COUNT,
            MetricType.TN_COUNT,
            MetricType.FP_COUNT,
            MetricType.FN_COUNT,
            MetricType.ACCURACY,
            MetricType.RECALL,
            MetricType.PRECISION,
            MetricType.F1,
        ]:
            # This was previously validated
            field: SDKField = values["field"]
            if field.type is FieldType.NUMERIC and values["threshold"] is None:
                raise ValueError(
                    "Must provide `threshold` parameter for numeric confusion-matrix metrics"
                )

            if (
                value
                in [
                    MetricType.RECALL,
                    MetricType.PRECISION,
                    MetricType.F1,
                ]
                and field.type is FieldType.CATEGORICAL
                and values["class_name"] is None
            ):
                raise ValueError(
                    "Must provide `average` parameter for multiclass confusion-matrix metrics"
                )

        return values


class MetricOverrideConfiguration(BaseModel):
    field: Optional[SDKField] = None
    average: Optional[AverageMethod] = None
    threshold: Optional[float] = None
    k: Optional[int] = None
    class_name: Optional[str] = None
    baseline: Optional[BaselineConfiguration] = None
    quantile: Optional[float] = None

    def _apply(self, original_metric: MetricConfiguration) -> MetricConfiguration:
        return MetricConfiguration(
            metric=original_metric.metric,
            field=self.field or original_metric.field,
            custom_metric=original_metric.custom_metric,
            average=self.average or original_metric.average,
            threshold=self.threshold or original_metric.threshold,
            k=self.k or original_metric.k,
            class_name=self.class_name or original_metric.class_name,
            baseline=self.baseline or original_metric.baseline,
            quantile=self.quantile or original_metric.quantile,
        )



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import FieldGroup
from aporia.sdk.metrics import AverageMethod, MetricType
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version
from aporia.sdk.widgets.base import (
    BaselineConfiguration,
    BaseWidget,
    DisplayOptions,
    FixedTimeframe,
    MetricConfiguration,
    MetricOverrideConfiguration,
    MetricParameters,
    VersionSelection,
    WidgetType,
)


class MetricWidgetDataOptionsAnnotations(BaseModel):
    phase: DatasetType
    metric: Union[MetricType, str]
    version: VersionSelection = VersionSelection()
    field: Optional[Dict] = None
    fieldCategory: Optional[FieldGroup] = None
    baseline: Optional[BaselineConfiguration] = None
    threshold: Optional[float] = None
    metricAtK: Optional[int] = None
    quantile: Optional[float] = None
    metricPerClass: Optional[str] = None
    average: Optional[AverageMethod] = None
    parameters: Optional[MetricParameters] = None
    isModelMetric: bool = True
    isCustomMetric: bool = False
    isAggregableMetric: bool = True
    isCodeBasedMetric: bool = False
    dataSegment: Optional[Dict] = None
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[
        Any
    ] = None  # TODO: For some reason, this is casting numerics to strings. Changed "Union[int, float, str]" to "Any"
    dataSegmentBucketName: Optional[str] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricWidgetDataOptionsFilters(BaseModel):
    annotations: MetricWidgetDataOptionsAnnotations
    timeframe: Optional[Union[str, FixedTimeframe]] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: MetricWidgetDataOptionsFilters


class MetricWidget(BaseWidget):
    type: Literal[WidgetType.METRIC] = WidgetType.METRIC
    dataOptions: List[MetricWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        metric: MetricConfiguration,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        phase: DatasetType = DatasetType.SERVING,
        version: Optional[Version] = None,
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
    ) -> "MetricWidget":
        if (segment is None) != (segment_value is None):
            raise ValueError(
                "For using segments, both segment and segment_value arguments are needed"
            )

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        return MetricWidget(
            _internal_state={"metric": metric},
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                MetricWidgetDataOptions(
                    filters=MetricWidgetDataOptionsFilters(
                        timeframe=timeframe,
                        annotations=MetricWidgetDataOptionsAnnotations(
                            phase=phase,
                            metric=metric._get_metric(),
                            version=version_selection,
                            field=metric._get_field(),
                            fieldCategory=metric._get_field_category(),
                            baseline=metric.baseline,
                            threshold=metric.threshold,
                            metricAtK=metric.k,
                            quantile=metric.quantile,
                            metricPerClass=metric.class_name,
                            average=metric.average,
                            parameters=metric._get_parameters(),
                            isModelMetric=metric._is_model_metric(),
                            isCustomMetric=metric._is_custom_metric(),
                            isAggregableMetric=metric._is_aggregable_metric(),
                            isCodeBasedMetric=metric._is_code_based_metric(),
                            dataSegment=segment.to_widget() if segment is not None else None,
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=segment_value if segment is not None else None,
                            dataSegmentBucketName=(
                                [
                                    bucket["name"]
                                    for bucket in segment.to_widget()["buckets"]
                                    if bucket["value"] == segment_value
                                ][0]
                                if segment is not None
                                else None
                            ),
                        ),
                    )
                )
            ],
        )

    def compare(
        self,
        phase: DatasetType = DatasetType.SERVING,
        version: Optional[Version] = None,
        timeframe: Optional[Union[str, FixedTimeframe]] = None,
        metric_overrides: MetricOverrideConfiguration = MetricOverrideConfiguration(),
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
    ) -> "MetricWidget":
        # TODO: Consider enforcing compare limits here
        original_metric: MetricConfiguration = self._internal_state["metric"]
        metric = metric_overrides._apply(original_metric)
        color = self.dataOptions[-1].display.color + 1

        version_selection = (
            VersionSelection(id=version.id) if version is not None else VersionSelection()
        )

        self.dataOptions.append(
            MetricWidgetDataOptions(
                display=DisplayOptions(color=color),
                filters=MetricWidgetDataOptionsFilters(
                    timeframe=timeframe,
                    annotations=MetricWidgetDataOptionsAnnotations(
                        phase=phase,
                        metric=metric._get_metric(),
                        version=version_selection,
                        field=metric._get_field(),
                        fieldCategory=metric._get_field_category(),
                        baseline=metric._get_baseline(),
                        threshold=metric._get_threshold(),
                        metricAtK=metric._get_k(),
                        quantile=metric._get_quantile(),
                        metricPerClass=metric._get_class(),
                        average=metric._get_average(),
                        parameters=metric._get_parameters(),
                        isModelMetric=metric._is_model_metric(),
                        isCustomMetric=metric._is_custom_metric(),
                        isAggregableMetric=metric._is_aggregable_metric(),
                        isCodeBasedMetric=metric._is_code_based_metric(),
                        dataSegment=segment.to_widget() if segment is not None else None,
                        dataSegmentId=segment.id if segment is not None else None,
                        dataSegmentName=segment.name if segment is not None else None,
                        dataSegmentValue=segment_value if segment is not None else None,
                        dataSegmentBucketName=(
                            [
                                bucket["name"]
                                for bucket in segment.to_widget()["buckets"]
                                if bucket["value"] == segment_value
                            ][0]
                            if segment is not None
                            else None
                        ),
                    ),
                ),
            ),
        )
        return self
